// ==UserScript==
// @name         Standings for HU & Hitachi NCCC 2018 JP
// @namespace    http://tampermonkey.net/
// @version      0.1.0
// @description  Show standings along the special rule
// @author       Hitachi Hokkaido University Labolatory
// @license      MIT License
// @include      *://atcoder.jp/contests/hokudai-hitachi2018/standings*
// @resource     tablesorter-css https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.1/css/theme.default.css
// @require      https://use.fontawesome.com/releases/v5.0.6/js/all.js
// @require      https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js
// @require      https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.1/js/jquery.tablesorter.js
// @require      https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.1/js/jquery.tablesorter.widgets.js
// @grant        GM_addStyle
// @grant        GM_getResourceText
// ==/UserScript==

(function() {
    'use strict';

    // 通常の順位表にするかどうか
    function getParam(name, url) {
        if (!url) url = window.location.href;
        name = name.replace(/[\[\]]/g, "\\$&");
        var regex = new RegExp("[?&]" + name + "(=([^&#]*)|&|#|$)"),
            results = regex.exec(url);
        if (!results) return null;
        if (!results[2]) return '';
        return decodeURIComponent(results[2].replace(/\+/g, " "));
    }

    var use_normal_standings_flag = (getParam('use_normal_standings', document.location.href) == 1);

    // .clearfix にボタンを追加
    var button_line = '<button type="button" class="btn' + (use_normal_standings_flag ? ' btn-primary' : '') + '" style="margin-right:10px;" onClick="location.href=\'./standings?use_normal_standings=0\'">日立北大基準の順位表にする</button>';
    button_line += '<button type="button" class="btn' + (use_normal_standings_flag ? '' : ' btn-primary') + '" onClick="location.href=\'./standings?use_normal_standings=1\'">通常の順位表にする</button>';
    $('.clearfix').append(button_line);

    if(use_normal_standings_flag) return;

    var tablesorter_css = GM_getResourceText ('tablesorter-css');
    GM_addStyle (tablesorter_css);

    // 未提出の扱い
    var NOSUB = 0;

    // 元のページャー・パネル・順位表 (の行) を削除
    $('ul.pagination').remove();
    $('span.pull-right').remove();
    $('#standings-panel-heading').remove();
    $('thead tr').remove();
    $('tbody tr').remove();

    // tablesorter の動作に必要な class を table タグに追加
    $('table').attr('id', 'standings-tablesorter');
    $('table').addClass('tablesorter');

    // 順位表データの取得
    var contest_id = 'hokudai-hitachi2018';
    var json_url = 'https://atcoder.jp/contests/' + contest_id + '/standings/json';
    $.getJSON(json_url, function(data) {
        // 問題 ID 一覧
        var problem_id = data.TaskInfo.map((e) => e.TaskScreenName);

        // thead を追加
        var th_line = '<th label="num" class="" style="width:4%;" data-name="rank" style="padding:0;">順位</th>';
        th_line += '<th label="nocase" class="standings-th-user" data-name="rating">ユーザ</th>';
        th_line += '<th label="num" class="no-break standings-result-th" style="width:140px;min-width:140px;" data-name="score">得点</th>';

        for(var x=0; x<problem_id.length; x++) {
            var alph = String.fromCharCode( 'A'.charCodeAt(0) + x );
            th_line += '<th label="num" class="standings-result-th" style="width:140px;min-width:140px;" data-name="' + problem_id[x] + '"><a class="standings-task-th" href="/contests/' + contest_id + '/tasks/' + problem_id[x] + '" target="_blank">' + alph + '</a></th>';
        }
        $('thead').append('<tr>' + th_line + '</tr>');

        // 各問題ごとの順位づけ (得点のみ考慮)
        problem_id.forEach(function(pid) {
            var sorted = data.StandingsData.map((e) => e.TaskResults.hasOwnProperty(pid) ? e.TaskResults[pid].Score : NOSUB).sort((a, b) => b - a);

            data.StandingsData.forEach(function(user) {
                // pid をメンバに持っていない場合は新しく作る
                if(!user.TaskResults.hasOwnProperty(pid)) {
                    user.TaskResults[pid] = {};
                    user.TaskResults[pid].rank = sorted.indexOf(NOSUB) + 1;
                }
                else {
                    user.TaskResults[pid].rank = sorted.indexOf(user.TaskResults[pid].Score) + 1;
                }
            });
        });

        // 総合順位用のデータを取得
        data.StandingsData.forEach(function(user) {
            var rank_prod       = 1;
            var rank_min        = Infinity;
            var last_submission = -1;

            problem_id.forEach(function(pid) {
                rank_prod *= user.TaskResults[pid].rank;
                rank_min = Math.min(rank_min, user.TaskResults[pid].rank);
                // 得点が正の場合のみ、最終提出時間を考慮
                if(user.TaskResults[pid].hasOwnProperty('Elapsed') && user.TaskResults[pid].Score > 0) {
                    last_submission = Math.max(last_submission, user.TaskResults[pid].Elapsed);
                }
            });

            user.rank_prod = rank_prod;
            user.rank_min = rank_min;
            user.last_submission = last_submission;
        });

        // 総合順位の指標でユーザーをソート
        // ソート以外でも比較するので比較関数を定義しておく
        function compareUser(a, b) {
            if(a.rank_prod != b.rank_prod) return a.rank_prod - b.rank_prod;
            if(a.rank_min != b.rank_min) return a.rank_min - b.rank_min;
            return a.last_submission - b.last_submission;
        }

        // レーティングから色を取得 (赤以上は全部赤に)
        function getRatingColor(user) {
            if(!user.hasOwnProperty('Rating') || user.Rating == 0) return 'user-unrated';
            var rating_color = [
                {'rating':  400, 'color': 'user-gray'},
                {'rating':  800, 'color': 'user-brown'},
                {'rating': 1200, 'color': 'user-green'},
                {'rating': 1600, 'color': 'user-cyan'},
                {'rating': 2000, 'color': 'user-blue'},
                {'rating': 2400, 'color': 'user-yellow'},
                {'rating': 2800, 'color': 'user-orange'}
            ];
            for(var k=0; k<rating_color.length; k++) {
                if(user.Rating < rating_color[k].rating) return rating_color[k].color;
            }
            return 'user-red';
        }

        // 提出時間をわかりやすい形で表示
        function getSubmissionTime(user_time) {
            var elapsed_time = user_time / 1000000000;
            var day  = Math.floor(elapsed_time / (24 * 60 * 60));
            var hour = ('00' + Math.floor(elapsed_time % (24 * 60 * 60) / (60 * 60))).slice(-2);
            var min  = ('00' + Math.floor(elapsed_time % (60 * 60) / (60))).slice(-2);
            var sec  = ('00' + elapsed_time % 60).slice(-2);
            var res  = (day === 0 ? '' : day + 'd ');
            return res + hour + ':' + min + ':' + sec;
        }

        function getSubmissionInfo(user, pid) {
            // 提出がない、または 0 点
            if(!user.TaskResults[pid].hasOwnProperty('Score') || user.TaskResults[pid].Score == 0) {
                return '-';
            }
            else {
                var actual_score = user.TaskResults[pid].Score / 100;
                return actual_score + ' / ' + getSubmissionTime(user.TaskResults[pid].Elapsed);
            }
        }

        data.StandingsData.sort(compareUser);

        var num_of_users = data.StandingsData.length;
        var current_rank = 1;
        for(var k=0; k<num_of_users; k++) {
            if(k == 0) {
                data.StandingsData[k].total_rank = current_rank;
            }
            else {
                // 前のユーザーと優劣がつかない場合は同順位
                // そうでなければ順位は変動
                if(compareUser(data.StandingsData[k-1], data.StandingsData[k]) !== 0) current_rank = k + 1;
                data.StandingsData[k].total_rank = current_rank;
            }

            // 順位表 (の行) を追加
            var user = data.StandingsData[k];
            var line = '<td class="standings-rank"><div class="extract-elem">' + user.total_rank + '</div></td>';
            line += '<td class="standings-username">';
            line += '<img src="//img.atcoder.jp/assets/flag/' + user.Country + '.png" class="img-flag-btn" data-code="' + user.Country + '"> ';
            line += '<a class="username" href="/users/' + user.UserScreenName + '">';
            line += '<span class="' + getRatingColor(user) + ' extract-elem">' + user.UserScreenName + '</span>';
            line += '</a>';
            line += '<span class="standings-user-btn">'
            line += ' <a href="/contests/' + contest_id + '/submissions?f.User=' + user.UserScreenName + '"><span class="glyphicon glyphicon-search black" aria-hidden="true" data-html="true" data-toggle="tooltip" title="" data-original-title="' + user.UserScreenName + 'さんの提出を見る"></span></a>';
            line += '</span>';
            line += '</td>';

            line += '<td class="standings-result">';
            line += '<p class="extract-elem"><span class="standings-score">' + user.rank_prod + '</span></p>';
            line += '<p>' + (user.last_submission < 0 ? '-' : getSubmissionTime(user.last_submission)) + '</p>';
            line += '</td>';

            problem_id.forEach(function(pid) {
                line += '<td class="standings-result">';
                line += '<p class="extract-elem"><span class="standings-ac">' + user.TaskResults[pid].rank + '</span></p>';
                line += '<p>' + getSubmissionInfo(user, pid) + '</p>';
                line += '</td>';
            });

            if(user.UserScreenName == userScreenName) {
                $('tbody').append('<tr class="info">'+line+'</tr>');
            } else {
                $('tbody').append('<tr>'+line+'</tr>');
            }
        }

        // 自分の近くに移動
        if ($('tr.info').offset()) {
            $("html, body").animate({ scrollTop: $('tr.info').offset().top - 200 }, 200);
        }

        var extractData = function(node) {
            var e = node.getElementsByClassName('extract-elem');
            return e[0].innerText;
        }

        $('document').ready(function() {
            $('#standings-tablesorter').tablesorter({
                textExtraction: extractData
            });
        });
    });
})();
